package be.iminds.ilabt.jfed.rspec.model.imutable_impl;

import be.iminds.ilabt.jfed.fedmon.webapi.service.json.Server;
import be.iminds.ilabt.jfed.lowlevel.resourceid.ResourceUrn;
import be.iminds.ilabt.jfed.lowlevel.testbed_info.TestbedInfoSource;
import be.iminds.ilabt.jfed.rspec.basic_model.BasicStringRspec;
import be.iminds.ilabt.jfed.rspec.model.LinkSetting;
import be.iminds.ilabt.jfed.rspec.model.RspecInterface;
import be.iminds.ilabt.jfed.rspec.model.RspecLink;
import be.iminds.ilabt.jfed.rspec.model.RspecNode;
import be.iminds.ilabt.jfed.rspec.parser.ExtraXml;
import be.iminds.ilabt.jfed.util.common.GeniUrn;
import be.iminds.ilabt.jfed.util.common.TextUtil;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import java.util.*;
import java.util.function.Function;
import java.util.function.Predicate;
import java.util.stream.Collectors;

public class ImmutableRspecLink implements RspecLink {
    private final static Logger LOG = LoggerFactory.getLogger(ImmutableRspecLink.class);

    protected final List<ExtraXml> extraXml;
    private final List<GeniUrn> componentManagerUrns;
    private final List<String> linkTypes;
    private final List<ImmutableRspecInterface> interfaces;
    private final List<ImmutableLinkSetting> linkSettings;
    private final String clientId;
    private final GeniUrn sliverId;
    private final String sharedLanName;
    @Nonnull private final ImmutableModelRspec modelRspec;

    private final List<String> autoSharedLanNames;

    private final Boolean vlanTagging;
    private final Integer vlanTag;
    private final Boolean noMacLearning;
    private final Boolean linkMultiplexing;
    private final Boolean bestEffort;
    private final Boolean interswitchDisallow;
    private final String noMacLearningValueString;

    private static <T> List<T> listInit(List<T> orig) {
        if (orig == null) return null;
        return Collections.unmodifiableList(new ArrayList<>(orig));
    }

    public ImmutableRspecLink(@Nonnull ImmutableModelRspec newModelRspec,
                              @Nonnull RspecLink orig) {
        this.modelRspec = newModelRspec;
        this.clientId = orig.getClientId();

        this.extraXml = listInit(orig.getExtraXml());
        this.sliverId = orig.getSliverId();
        this.componentManagerUrns = listInit(orig.getComponentManagerUrns());
        this.linkTypes = listInit(orig.getLinkTypes());

        this.sharedLanName = orig.getSharedLan();
        this.autoSharedLanNames = listInit(orig.getAutoShareLanNames());
        this.vlanTagging = orig.hasVlanTaggingTag() ? orig.isVlanTaggingEnabled() : null;
        this.vlanTag = orig.getVlanTag();
        this.noMacLearning = orig.hasNoMacLearningTag() ? orig.isNoMacLearning() : null;
        this.noMacLearningValueString = orig.hasNoMacLearningTag() ? orig.getNoMacLeaningValueString() : null;
        this.bestEffort = orig.hasBestEffortTag() ? orig.isBestEffortEnabled() : null;
        this.linkMultiplexing = orig.hasLinkMultiplexingTag() ? orig.isLinkMultiplexingEnabled() : null;
        this.interswitchDisallow = orig.hasInterswitchDisallowTag() ? orig.isInterswitchDisallowed() : null;

        List<ImmutableRspecInterface> copiedInterfaces = new ArrayList<>();
        for (RspecInterface origInterface : orig.getInterfaces()) {
            ImmutableRspecInterface copiedIface = newModelRspec.getInterfaceByClientId(origInterface.getClientId());
            if (copiedIface != null) {
                //use already copied iface, but bind to it.
                copiedIface.bindLink(this);
            } else {
                //clone iface
                copiedIface = new ImmutableRspecInterface(this.getRspec(), null, this, origInterface);
            }
            copiedInterfaces.add(copiedIface);
        }
        this.interfaces = Collections.unmodifiableList(copiedInterfaces);

        List<ImmutableLinkSetting> copiedLinkSetting = new ArrayList<>();
        //then copy linksettings
        for (LinkSetting origLinkSetting : orig.getLinkSettings()) {
            ImmutableRspecInterface fromIface = getInterfaceByClientId(origLinkSetting.getFromIface().getClientId());
            ImmutableRspecInterface toIface = getInterfaceByClientId(origLinkSetting.getToIface().getClientId());
            assert fromIface != null;
            assert toIface != null;
            copiedLinkSetting.add(new ImmutableLinkSetting(fromIface, toIface, origLinkSetting));
        }
        linkSettings = Collections.unmodifiableList(copiedLinkSetting);
    }

    @Override
    public void overwritePropertiesWith(@Nonnull RspecLink otherLink, boolean overwriteWithNulloverwriteNonAuthoratativeComponentManagers) {
        throw new UnsupportedOperationException();
    }

    @Override
    @Nonnull
    public ImmutableModelRspec getRspec() {
        return modelRspec;
    }

    @Nonnull
    @Override
    public String getUniqueId() {
        return BasicStringRspec.makeUniqueLinkId(clientId, null, getComponentManagerUrnStrings());
    }

    @Nonnull
    public List<String> getComponentManagerUrnStrings() {
        return componentManagerUrns.stream().map(ResourceUrn::getValue).collect(Collectors.toList());
    }

    @Override
    @Nullable
    public String getClientId() {
        return clientId;
    }

    @Override
    public void setClientId(@Nullable String value) {
        throw new UnsupportedOperationException();
    }

    @Nonnull
    @Override
    public List<GeniUrn> getComponentManagerUrns() {
        return componentManagerUrns;
    }

    @Nonnull
    @Override
    public List<String> getLinkTypes() {
        return linkTypes;
    }

    @Nonnull
    @Override
    public List<? extends LinkSetting> getLinkSettings() {
        return linkSettings;
    }

    @Override
    @Nullable
    public ImmutableLinkSetting getLinkSetting(@Nonnull RspecInterface fromIface, @Nonnull RspecInterface toIface) {
        assert fromIface != null;
        assert fromIface instanceof ImmutableRspecInterface;
        assert toIface != null;
        assert toIface instanceof ImmutableRspecInterface;
        //this one only makes sense within the same modelRspec
        assert fromIface.getRspec() == getRspec();
        assert toIface.getRspec() == getRspec();

        for (ImmutableLinkSetting ls : linkSettings) {
            if (ls.getFromIface() == fromIface && ls.getToIface() == toIface)
                return ls;
        }

        return null;
    }

    @Override
    @Nullable
    public ImmutableLinkSetting getLinkSettingByClientIds(@Nonnull String fromIfaceName, @Nonnull String toIfaceName) {
        for (ImmutableLinkSetting ls : linkSettings) {
            if (Objects.equals(ls.getFromIface().getClientId(), fromIfaceName) &&
                    Objects.equals(ls.getToIface().getClientId(), toIfaceName))
                return ls;
        }
        return null;
    }

    @Override
    @Nullable
    public ImmutableLinkSetting getLinkSettingByUniqueIds(@Nonnull String fromIfaceName, @Nonnull String toIfaceName) {
        for (ImmutableLinkSetting ls : linkSettings) {
            if (Objects.equals(ls.getFromIface().getUniqueId(), fromIfaceName) &&
                    Objects.equals(ls.getToIface().getUniqueId(), toIfaceName))
                return ls;
        }
        return null;
    }

    @Override
    public boolean isImpaired() {
        synchronized (linkSettings) {
            for (ImmutableLinkSetting ls : linkSettings) {
                if (ls.isActive())
                    return true;
            }
            return false;
        }
    }

    protected boolean addInterface(@Nonnull RspecInterface iface) {
        throw new UnsupportedOperationException();
    }

    protected boolean removeInterface(@Nonnull RspecInterface iface) {
        throw new UnsupportedOperationException();
    }

    @Nonnull
    @Override
    public List<ImmutableRspecInterface> getInterfaces() {
        return interfaces;
    }

    @Override
    @Nullable
    public ImmutableRspecInterface getInterfaceForNode(@Nonnull RspecNode node) {
        assert node instanceof ImmutableRspecNode;
        for (ImmutableRspecInterface iface : interfaces) {
            if (iface.getNode() == node)
                return iface;
        }
        return null;
    }

    @Override
    @Nullable
    public ImmutableRspecInterface getInterfaceByClientId(@Nonnull String clientId) {
        for (ImmutableRspecInterface iface : interfaces) {
            if (Objects.equals(iface.getClientId(), clientId))
                return iface;
        }
        return null;
    }

    @Override
    @Nullable
    public ImmutableRspecInterface getInterfaceByUniqueId(@Nonnull String uniqueId) {
        for (ImmutableRspecInterface iface : interfaces) {
            if (Objects.equals(iface.getUniqueId(), uniqueId))
                return iface;
        }
        return null;
    }

    @Override
    @Nullable
    public GeniUrn getSliverId() {
        return sliverId;
    }

    @Override
    public void setSliverId(@Nullable GeniUrn sliverId) {
        throw new UnsupportedOperationException();
    }

    @Override
    public void setSliverId(@Nullable String sliverId) {
        throw new UnsupportedOperationException();
    }

    @Override
    public boolean isStitched(@Nullable TestbedInfoSource testbedInfoSource) {
        if (getComponentManagerUrns().size() < 2) return false;

        //only links with type "lan" or links without type can be stitching
        if (getLinkTypes().size() > 1 ||
                (getLinkTypes().size() == 1 && !(getLinkTypes().get(0).equalsIgnoreCase("lan") || getLinkTypes().get(0).equalsIgnoreCase("vlan")))
                )
            return false;

        if (testbedInfoSource == null) return true;

        Server firstServer = null;
        for (GeniUrn urn : getComponentManagerUrns()) {
            //TODO this is actually more complex, depends on exogeni setting setc
            Server server =
                    testbedInfoSource.getByUrn(urn, TestbedInfoSource.SubAuthMatchAllowed.ALLOW_TOPLEVEL, TestbedInfoSource.SubAuthMatchPreference.PREFER_TOPLEVEL);
            if (server == null) {
                LOG.warn("Authority not found for " + urn + " during stitch check -> falling back to assuming link is stitched");
                return true; //fall back to assuming it is stitched, if any authority is not found
            }
            if (firstServer == null)
                firstServer = server;
            else {
                if (!Objects.equals(firstServer.getDefaultComponentManagerUrn(), server.getDefaultComponentManagerUrn()))
                    return true; //authorities have different URN, so stitched link
            }
        }
        return false;
    }

    @Nonnull
    @Override
    public List<ExtraXml> getExtraXml() {
        return extraXml;
    }

    @Override
    @Nullable
    public String getSharedLan() {
        return sharedLanName;
    }

    @Override
    public void setSharedLan(@Nullable String sharedLanName) {
        throw new UnsupportedOperationException();
    }


    @Override
    public boolean isVlanTaggingEnabled() {
        return vlanTagging != null && vlanTagging;
    }

    @Override
    public boolean hasVlanTaggingTag() {
        return vlanTagging != null;
    }

    @Override
    public void setVlanTagging(@Nullable Boolean value) {
        throw new UnsupportedOperationException();
    }

    @Override
    @Nullable
    public Integer getVlanTag() {
        return vlanTag;
    }

    @Override
    public void setVlanTag(@Nullable Integer vlanTag) {
        throw new UnsupportedOperationException();
    }

    @Override
    public boolean isNoMacLearning() {
        return noMacLearning != null && noMacLearning;
    }

    @Override
    public boolean hasNoMacLearningTag() {
        return noMacLearning != null;
    }

    @Override
    public void setNoMacLearning(@Nullable Boolean value) {
        throw new UnsupportedOperationException();
    }

    @Override
    @Nullable
    public String getNoMacLeaningValueString() {
        if (noMacLearning == null) return null;
        return noMacLearningValueString;
    }

    @Override
    public void setNoMacLearning(@Nullable String value) {
        throw new UnsupportedOperationException();
    }

    @Override
    public boolean isBestEffortEnabled() {
        return bestEffort != null && bestEffort;
    }

    @Override
    public boolean hasBestEffortTag() {
        return bestEffort != null;
    }

    @Override
    public void setBestEffort(@Nullable Boolean value) {
        throw new UnsupportedOperationException();
    }

    @Override
    public boolean isLinkMultiplexingEnabled() {
        return linkMultiplexing != null && linkMultiplexing;
    }

    @Override
    public boolean hasLinkMultiplexingTag() {
        return linkMultiplexing != null;
    }

    @Override
    public void setLinkMultiplexing(@Nullable Boolean value) {
        throw new UnsupportedOperationException();
    }


    @Override
    public boolean hasAutoShareLan() {
        return !autoSharedLanNames.isEmpty();
    }

    @Nonnull
    @Override
    public List<String> getAutoShareLanNames() {
        return autoSharedLanNames;
    }

    @Override
    public boolean isInterswitchDisallowed() {
        return interswitchDisallow != null && interswitchDisallow;
    }

    @Override
    public boolean hasInterswitchDisallowTag() {
        return interswitchDisallow != null;
    }

    @Override
    public void setInterswitchDisallow(@Nullable Boolean value) {
        throw new UnsupportedOperationException();
    }
}
